(* Blocks of tableaux. *)
open Core.Std open Constraint open Looping open Purity open Schema open Satisfiability open Literal

module Block =
  struct
    (* Blocks are just the pair of a set of schemas and of a constraint + various precomputed information. *)
    type t = Schema.t list * Constraint.t

    (* Constructors & Selectors *)
    let of_schs_and_cstr ~cstr ~schs = (schs,cstr)
    let schs_of = Tuple2.get1
    let cstr_of = Tuple2.get2
    let is_empty (schs,_) = match schs with [] -> true | _ -> false

    (* Print stuff *)
    let to_string (schs,cstr) = (Schema.list_to_string schs) ^ "  " ^ (Constraint.to_string ~parenthesized:true cstr)

    (* Maximum upper bound of all schemas *)
    let max_bound (schs,_) = List.fold ~f:(fun current_max -> Schema.max_bound ~acc:current_max) ~init:None schs

    (* True if a block contains a schema containing iterations *)
    let contains_iterations (schs,_) = List.exists schs ~f:Schema.contains_iterations

    (* Purity interface *)
    type purable = t

    (* Just apply purity_add to each schema *)
    let purity_add (schs,_) ?(range=None) ~pur = List.fold schs ~f:(fun pur -> Schema.purity_add ~range:None ~pur) ~init:pur

    (* Removes all the pure literals of the set *)
    let purify (schs,cstr) ~cstr ~pur =
      let m,non_pures = List.fold schs ~init:(Modified.Unmodified,[]) ~f:(fun acc sch ->
        match acc,Schema.purify ~cstr ~pur sch with
        |(m1,schs'),(m2,(`Top:Schema.t)) -> Modified.max m1 m2,schs'
        |(m1,schs'),(m2,sch)             -> Modified.max m1 m2,sch::schs')
      in
      m,(non_pures,cstr)

    (* Computes the purity of the whole block *)
    let purity = purity_add ~range:None ~pur:Purity.empty

    (* Normalize the block.
     * Approx. definition:
     *   Let P_e or Op_i=f..e be the first elt of a list of schemata [S] s.t. e contains the parameter n.
     *   If there is no such e then [S] is already normalized.
     *   Else the normalized of [S] is the list which is just the same as [S] except that every occurrence is shifted by n-e.
     *   So after normalization e is equal to n.
     *   TODO: also modify the constraints (useless for now as we don't use them in looping)
     *)
    let normalize (schs,cstr) =
      let rec normalize' = function
        |[]                 -> []
        |sch::schs' as schs -> 
            (match Schema.expr_involving_param sch with
            |None   -> sch::(normalize' schs')
            |Some e -> 
                (match Indexes.get_var e with
                |None     -> assert false
                |Some var -> List.map schs ~f:(Schema.substitute ~var ~by:(Indexes.inverse e))
                )
            )
      in
      (normalize' schs,cstr)

    (* Comparison *)
    let rec schs_equal schs1 schs2 =
      match schs1,schs2 with
      |[],[]                     -> true
      |sch1::schs1',sch2::schs2' -> Schema.equal sch1 sch2 && schs_equal schs1' schs2'
      |_                         -> false
    let equal (schs1,cstr1) (schs2,cstr2) = Constraint.equal cstr1 cstr2 && schs_equal schs1 schs2

    (* Removal of duplicate schemata *)
    let dedup (schs,cstr) =
      let sorted_schs = List.sort ~cmp:(fun x y -> Schema.total_compare y x) schs in
      let rec dedup_schs ~acc = function
        |[]             -> acc
        |[sch]          -> sch::acc
        |sch1::sch2::schs -> if Schema.equal sch1 sch2 then dedup_schs (sch2::schs) ~acc else dedup_schs (sch2::schs) ~acc:(sch1::acc)
      in 
      (dedup_schs sorted_schs ~acc:[],cstr)

    (* Generic strategies to apply some given rules *)
    type branch_update =
      |Replace_with_schema                of Schema.t
      |Replace_with_schemas               of Schema.t * Schema.t
      |Replace_with_schema_add_constraint of Schema.t * Constraint.t
      |Add_constraint                     of Constraint.t

    type branching =
      |Go_on
      |Close
      |Update of branch_update
      |Branch of branch_update * branch_update

    let apply_rules_with_closure (schs,cstr) ~rules ~k =
      let rec traverse ~schs ~cstr ~acc ~litmap =
        match schs with
        |[]                                                        -> k (acc,cstr)
        |((`Lit l):Schema.t)::_ when ClosureSet.closure_applies litmap ~key:l -> NoModelFound
        |sch::schs'                                                -> 
            let update ~k = function
              |Replace_with_schema sch'                       -> k (traverse ~schs:(sch'::schs') ~cstr ~acc ~litmap)
              |Replace_with_schemas(sch1,sch2)                -> k (traverse ~schs:(sch1::sch2::schs') ~cstr ~acc ~litmap)
              |Replace_with_schema_add_constraint(sch',cstr') -> k (traverse ~schs:(sch'::schs') ~cstr:(bounds_conjunction cstr cstr') ~acc ~litmap)
              |Add_constraint cstr'                           -> k (traverse ~schs:schs' ~cstr:(bounds_conjunction cstr cstr') ~acc ~litmap)
            in
            (match rules ~sch ~cstr with
            |Go_on         -> traverse ~schs:schs' ~cstr ~acc:(sch::acc) ~litmap:(
                match sch with ((`Lit l):Schema.t) -> ClosureSet.add litmap ~key:l | _ -> litmap)
            |Close         -> NoModelFound
            |Update u      -> update ~k:ident u
            |Branch(u1,u2) -> let second_branch = function ModelFound -> ModelFound | NoModelFound -> update ~k:ident u2 in update ~k:second_branch u1)
      in
      traverse ~schs ~cstr ~acc:[] ~litmap:(ClosureSet.empty ~cstr)

    let apply_rules (schs,cstr) ~rules ~k =
      let rec traverse ~schs ~cstr ~acc =
        match schs with
        |[]                                                       -> k (acc,cstr)
        |sch::schs'                                               -> 
            let update ~k = function
              |Replace_with_schema sch'                       -> k (traverse ~schs:(sch'::schs') ~cstr ~acc)
              |Replace_with_schemas(sch1,sch2)                -> k (traverse ~schs:(sch1::sch2::schs') ~cstr ~acc)
              |Replace_with_schema_add_constraint(sch',cstr') -> k (traverse ~schs:(sch'::schs') ~cstr:(bounds_conjunction cstr cstr') ~acc)
              |Add_constraint cstr'                           -> k (traverse ~schs:schs' ~cstr:(bounds_conjunction cstr cstr') ~acc)
            in
            (match rules ~sch ~cstr with
            |Go_on         -> traverse ~schs:schs' ~cstr ~acc:(sch::acc)
            |Close         -> NoModelFound
            |Update u      -> update ~k:ident u
            |Branch(u1,u2) -> let second_branch = function ModelFound -> ModelFound | NoModelFound -> update ~k:ident u2 in update ~k:second_branch u1)
      in
      traverse ~schs ~cstr ~acc:[]
      
  end

include Block

open Looping 

(* Finally we can construct the trie from all the maps of other data strutures.
 *
 * Important note: blocks are not considered as sets but indeed as lists.
 * Termination property still holds and this is much more efficient (and readable) this way.
 *)
module LoopingMap =
  struct
    module SchMap = Schema.LoopingMap
    type loopable = Block.t
    type 'a t = ('a t) SchMap.t * 'a option
    (* TOTRY:
     * - integrate constraint? i.e. ('a * Constraint.t) option; I think it's useless
     * - add a constructor Node of ... : allow to get rid of -rectypes; maybe there's a lost in efficiency 
     *)
    type param = unit

    let empty = (SchMap.empty,None)
    
    let add m ~key ~data =
      let rec add' (schmap,current_data) ~schs ~data =
        (match schs with
        |[]         -> schmap,(Some data)
        |sch::schs' -> (SchMap.add schmap ~key:sch ~data:(add' (Option.value (SchMap.find schmap ~key:sch) ~default:empty) ~schs:schs' ~data)), current_data)
      in
      let (schs,_) = normalize key in
      add' m ~schs ~data

    let find m ~key =
      let rec find' (m,data) ~schs =
        match schs with
        |[]         -> data
        |sch::schs' ->
            (match SchMap.find m ~key:sch with
            |None    -> None
            |Some m' -> find' m' ~schs:schs')
      in
      let (schs,_) = normalize key in
      find' m ~schs
  end
