(* Formulae. *)
open Core.Std

(* Currently the implementation is just the same as for schemata except for the 'It' constructor.
 *)
type operator = And | Or | Xor | Equiv | Imply
type t =
  |Lit of Literal.t
  |Op  of operator * t * t
  |Top
  |Bot

(* Constructors *)
let bot                   = Bot
let top                   = Top
let lit l                 = Lit(l)
let conj sch1 sch2        = Op(And,sch1,sch2)
let disj sch1 sch2        = Op(Or,sch1,sch2)
let xor sch1 sch2         = Op(Xor,sch1,sch2)
let equiv sch1 sch2       = Op(Equiv,sch1,sch2)
let imply sch1 sch2       = Op(Imply,sch1,sch2)

(* Computes the negation in negative normal form of a whole schema. *)
let rec neg = function
  |Top                  -> Bot
  |Bot                  -> Top
  |Lit(l)               -> Lit(Literal.opposite l)
  |Op(op, sch1, sch2)   -> neg_operator op sch1 sch2

and neg_operator op sch1 sch2 = 
  match op with
    |And   -> Op(Or   , neg sch1, neg sch2)
    |Or    -> Op(And  , neg sch1, neg sch2)
    |Xor   -> Op(Equiv, sch1, sch2)
    |Equiv -> Op(Xor  , sch1, sch2)
    |Imply -> Op(And  , sch1, neg sch2)

(* Canonize a formula i.e. eliminates redundant symbols (Xor, Equiv, Imply) *)
let canonized = function
  |Op(Xor,f1,f2)   -> Op(And,Op(Or,neg f1,neg f2),Op(Or,f1,f2)) 
  |Op(Equiv,f1,f2) -> Op(And,Op(Or,neg f1,f2),Op(Or,f1,neg f2)) 
  |Op(Imply,f1,f2) -> Op(Or,neg f1,f2) 
  |x               -> x

(* Conjonctive Normal Form  *)
exception NotInCNF

  (* Computes the conjunctive normal form (naive, no sharing) *)
let rec to_cnf f = 
  let rec distr f1 f2 =
    match f1,f2 with
    |Op(And,x2,x3),y -> Op(And, distr x2 y, distr x3 y)
    |x,Op(And,y2,y3) -> Op(And, distr x y2, distr x y3)
    |x,y             -> Op(Or,x,y)
  in 
  match canonized f with
  |Op(And,x,y) -> Op(And,to_cnf x, to_cnf y)
  |Op(Or,x,y)  -> distr (to_cnf x) (to_cnf y)
  |f'          -> f'

(* DIMACS generation *)

 (* Edds a mapping atom -> number to a given map.
  * The input ~acc is an object containing the map and the previous maximum integer assigned to an atom.
  * The use of objects here is for my pure pleasure. But completely unnecessary.
  *)
let add_var ~acc (prop,idx) = 
  if not (Map.mem acc#map (prop,idx)) then
      let max' = acc#max+1 in
      object
        method map=Map.add acc#map ~key:(prop, idx) ~data:max'
        method max=max' 
      end
  else acc

  (* Browse the formula to generate a mapping atom -> number *)
let rec vars_to_ints ?(acc=object method map=Map.empty method max=0 end) = function
  |Top | Bot   -> acc
  |Lit(l)      -> add_var ~acc (Literal.atom_of l)
  |Op(_,f1,f2) -> vars_to_ints ~acc:(vars_to_ints ~acc f1) f2

  (* If a literal is negative we prepend a '-' else nothing *)
let sgn_to_dimacs l = if Literal.is_pos l then "" else "-"

  (* Computes the string corresponding to one clause *)
let rec clause_to_dimacs ~map = function
  |Bot          -> ""
  |Lit(l)       -> (sgn_to_dimacs l) ^ (string_of_int (Map.find_exn map (Literal.atom_of l)))
  |Op(Or,f1,f2) -> (clause_to_dimacs ~map f1) ^ " " ^ (clause_to_dimacs ~map f2)
  |_            -> raise NotInCNF

  (* Computes a list of strings, each string being a clause in DIMACS format.
   * The use of a list in first place allows to compute easily the number of clauses in the formula.
   *)
let rec instances ~map ?(acc=[]) = function
  |Top           -> acc
  |Op(And,f1,f2) -> instances ~map ~acc:(instances ~map ~acc f1) f2
  |c             -> (clause_to_dimacs ~map c)::acc

  (* Generic introduction to the generated file *)
let dimacs_intro = "c Generated by sch2cnf.\nc\n"

  (* Computes the whole thing:
   * - intro
   * - description of the instance (nb of variables, nb of clauses)
   * - list of clauses
   *)
let to_dimacs f =
  let vars_to_ints = vars_to_ints f in
  let instances    = instances ~map:vars_to_ints#map (to_cnf f) in
  dimacs_intro ^ "p cnf " ^ (string_of_int vars_to_ints#max) ^ " " ^ (string_of_int (List.length instances)) ^ "\n" ^ (String.concat ~sep:" 0\n" instances)

(* Print stuff *)
let operator_to_string = function And -> "/\\"     | Or -> "\\/"   | Xor -> "(+)"     | Equiv -> "<->"              | Imply -> "->"
let operator_to_tex    = function And -> "\\wedge" | Or -> "\\vee" | Xor -> "\\oplus" | Equiv -> "\\Leftrightarrow" | Imply -> "\\Rightarrow"

let to_parenthesized ~f:to_string s =
  match s with
  |Lit(_) | Top | Bot -> to_string s
  |_                  -> "(" ^ (to_string s) ^ ")"

let rec to_string = function
  |Top          -> "True"
  |Bot          -> "False"
  |Lit(l)       -> Literal.to_string l
  |Op(op,s1,s2) -> (to_parenthesized ~f:to_string s1) ^ " " ^ (operator_to_string op) ^ " " ^ (to_parenthesized ~f:to_string s2)

let print        = print_endline <|| to_string
let print_cnf    = print         <|| to_cnf
let print_dimacs = print_endline <|| to_dimacs

