open Core.Std open OUnit open Indexes

let raises_exception f =
  try f (); false
  with _ -> true

module Values = 
  struct
    let vari      = Variable.of_string "i"
    let varv      = Variable.of_string "v"
    let varj      = Variable.of_string "j"
    let varn      = Variable.of_string "n"
    let varn'     = Variable.of_string "n'"
    let termvari  = of_var vari
    let termvarj  = of_var varj
    let termvarv  = of_var varv
    let termvarn  = of_var varn
    let termvarn' = of_var varn'
    let int1      = of_num 1
    let int2      = of_num 2
    let int0      = of_num 0
    let vplus1    = add termvarv 1
    let vplus2    = add termvarv 2
    let iplus1    = add termvari 1
    let iplus0    = add termvari 0
    let iminus1   = add termvari (-1)
    let iminus2   = add termvari (-2)
    let iplus2    = add termvari 2
    let nplus1    = add termvarn 1
    let nplus2    = add termvarn 2
    let nplus0    = add termvarn 0
    let nminus1   = add termvarn (-1)
    let nminus2   = add termvarn (-2)
    let minus1    = of_num (-1)
    let n'plus1   = add termvarn' 1
    let n'plus2   = add termvarn' 2
    let nplus3    = add termvarn 3
  end

open Values

let test =
  "termes" >:::
    ["to_string" >::
      (fun () ->
        "variable" @? (to_string termvarv = "v");
        "addition" @? (to_string vplus1   = "v+1");
        "int"      @? (to_string int1     = "1");
      );
    "equal" >::
      (fun () ->
        "1 1"     @? (equal int1 int1);
        "2 2"     @? (equal int2 int2);
        "n n"     @? (equal termvarn termvarn);
        "n+1 n+1" @? (equal nplus1 nplus1);
        "1 2"     @? (not (equal int1 int2));
        "1 n"     @? (not (equal int1 termvarn));
        "1 n+1"   @? (not (equal int1 nplus1));
        "2 1"     @? (not (equal int2 int1));
        "n 1"     @? (not (equal termvarn int1));
        "n+1 1"   @? (not (equal nplus1 int1));
        "n 2"     @? (not (equal termvarn int2));
        "n v"     @? (not (equal termvarn termvarv));
        "n n+1"   @? (not (equal termvarn nplus1));
      );
    "to_tex" >::
      (fun () ->
        "variable" @? (to_tex termvarv = "v");
        "addition" @? (to_tex vplus1   = "v+1");
        "int"      @? (to_tex int1     = "1");
      );
    "substitute" >::
      (fun () ->
        "intvar"    @? (substitute ~var:vari ~by:int1 termvari     = int1);
        "intint"    @? (substitute ~var:vari ~by:int1 int1         = int1);
        "intplus"   @? (substitute ~var:vari ~by:int1 iplus1       = int2);
        "varvar"    @? (substitute ~var:vari ~by:termvarv termvari = of_var varv);
        "varint"    @? (substitute ~var:vari ~by:termvarv int1     = int1);
        "varplus"   @? (substitute ~var:vari ~by:termvarv iplus1   = vplus1);
        "plusvar"   @? (substitute ~var:vari ~by:vplus2 termvari   = vplus2);
        "plusint"   @? (substitute ~var:vari ~by:vplus2 int1       = int1);
        "plusplus"  @? (substitute ~var:vari ~by:vplus2 iplus1     = add termvarv 3);
        "identvar"  @? (substitute ~var:varj ~by:vplus2 termvari   = of_var vari);
        "plusmoins" @? (substitute ~var:vari ~by:vplus2 iminus1    = vplus1);
        "zero"      @? (substitute ~var:vari ~by:vplus2 iminus2    = of_var varv);
      );
    "max" >::
      (fun () ->
        "n vs n+1"   @? (max termvarn nplus1  = nplus1);
        "n vs n+2"   @? (max termvarn nplus2  = nplus2);
        "n-1 vs n"   @? (max nminus1 termvarn = termvarn);
        "n-2 vs n"   @? (max nminus2 termvarn = termvarn);
        "n-1 vs n+1" @? (max nminus1 nplus1   = nplus1);
        "n-2 vs n+1" @? (max nminus2 nplus1   = nplus1);
        "n-2 vs n+2" @? (max nminus2 nplus2   = nplus2);
        "n+2 vs n+2" @? (max nplus2 nplus2    = nplus2);
        "n-2 vs n-1" @? (max nminus2 nminus1  = nminus1);
        "n vs 1"     @? raises_exception (fun () -> max termvarn int1);
        "1 vs n"     @? raises_exception (fun () -> max int1 termvarn);
        "1 vs n"     @? raises_exception (fun () -> max int1 nplus1);
        "n vs 1"     @? raises_exception (fun () -> max nplus2 int1);
        "n vs n"     @? (max termvarn termvarn = termvarn);
        "n vs j"     @? raises_exception (fun () -> max (of_var varj) termvarn);
        "j vs n"     @? raises_exception (fun () -> max termvarn (of_var varj));
        "n vs j+1"   @? raises_exception (fun () -> max termvarn (add (of_var varj) 1));
        "n+1 vs j+1" @? raises_exception (fun () -> max nplus1 (add (of_var varj) 1));
      );
    "add" >::
      (fun () ->
        "n+1"   @? (add termvari 1  = iplus1);
        "n+2"   @? (add termvari 2  = (add termvari 2));
        "n+0"   @? (add termvari 0  = termvari);
        "n+1+1" @? (add nplus1 1    = nplus2);
        "n+1+2" @? (add nplus1 2    = (add termvarn 3));
        "n+1+0" @? (add nplus1 0    = nplus1);
        "n-1+1" @? (add nminus1 1   = termvarn);
        "n-1+2" @? (add nminus1 2   = nplus1);
        "n-1+0" @? (add nminus1 0   = nminus1);
        "n+2-1" @? (add nplus2 (-1) = nplus1);
        "n+2-2" @? (add nplus2 (-2) = termvarn);
      );
    "get_var" >::
      (fun () ->
        "n"   @? (get_var termvari = (Some vari));
        "n+1" @? (get_var nplus1   = (Some varn));
        "i+1" @? (get_var iplus1   = (Some vari));
        "i-1" @? (get_var iminus1  = (Some vari));
        "1"   @? (get_var int1     = None);
      );
    "looping map" >:::
      let empty_set = LoopingMap.empty in
      let nplus1_set = LoopingMap.add  empty_set ~key:nplus1 ~data:() in
      let nminus1_set = LoopingMap.add empty_set ~key:nminus1 ~data:() in
      let n_set = LoopingMap.add       empty_set ~key:termvarn ~data:() in
      let int1_set = LoopingMap.add    empty_set ~key:int1 ~data:() in
      let int2_set = LoopingMap.add    empty_set ~key:int2 ~data:() in
      [
        "loops_among" >::
          (fun () ->
            "n   -> n"   @? (Option.is_some (LoopingMap.find n_set ~key:termvarn));
            "n+1 -> n+1" @? (Option.is_some (LoopingMap.find nplus1_set ~key:nplus1));
            "n-1 -> n-1" @? (Option.is_some (LoopingMap.find nminus1_set ~key:nminus1));
            "1   -> 1"   @? (Option.is_some (LoopingMap.find int1_set ~key:int1));
            "2   -> 2"   @? (Option.is_some (LoopingMap.find int2_set ~key:int2));
          );
     ]
    ]
